/*
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdinc.h"
#include "DCPlusPlus.h"

#include "QueueManager.h"

#include "ConnectionManager.h"
#include "SearchManager.h"
#include "ClientManager.h"
#include "DownloadManager.h"
#include "ShareManager.h"
#include "LogManager.h"
#include "ResourceManager.h"

#include "UserConnection.h"
#include "SimpleXML.h"
#include "StringTokenizer.h"
#include "DirectoryListing.h"

#include <limits>

#ifndef _WIN32
#include <sys/types.h>
#include <dirent.h>
#include <fnmatch.h>
#endif

const string QueueManager::USER_LIST_NAME = "MyList.DcLst";

namespace {
	const string TEMP_EXTENSION = ".tdc";

	string getTempName(const string& aFileName, const TTHValue* aRoot) {
		string tmp(aFileName);
		if(aRoot != NULL) {
			tmp += "." + aRoot->toBase32();
		}
		tmp += TEMP_EXTENSION;
		return tmp;
	}
}

const string& QueueItem::getTempTarget() {
	if(!isSet(QueueItem::FLAG_USER_LIST) && tempTarget.empty() && File::getSize(getTarget()) == -1) {
		if(!SETTING(TEMP_DOWNLOAD_DIRECTORY).empty()) {
#ifdef _WIN32
			::StringMap sm;
			if(target.length() >= 3 && target[1] == ':' && target[2] == '\\')
				sm["targetdrive"] = target.substr(0, 3);
			else
				sm["targetdrive"] = Util::getConfigPath().substr(0, 3);
			setTempTarget(Util::formatParams(SETTING(TEMP_DOWNLOAD_DIRECTORY), sm, false) + getTempName(getTargetFileName(), getTTH()));
#else //_WIN32
			setTempTarget(SETTING(TEMP_DOWNLOAD_DIRECTORY) + getTempName(getTargetFileName(), getTTH()));
#endif //_WIN32
		} else {
			setTempTarget(Util::getFilePath(getTarget()) + getTempName(getTargetFileName(), getTTH()));
		}
	}
	return tempTarget;
}

QueueItem* QueueManager::FileQueue::add(const string& aTarget, int64_t aSize,
						  Flags<uint16_t>::MaskType aFlags, QueueItem::Priority p, const string& aTempTarget,
						  int64_t aDownloadedBytes, time_t aAdded, const TTHValue* root) throw(QueueException, FileException) {
	if(aFlags & QueueItem::FLAG_USER_LIST) {
		p = QueueItem::FILELIST;
	} else if(p == QueueItem::DEFAULT) {
		if(aSize <= SETTING(PRIO_HIGHEST_SIZE)*1024) {
			p = QueueItem::HIGHEST;
		} else if(aSize <= SETTING(PRIO_HIGH_SIZE)*1024) {
			p = QueueItem::HIGH;
		} else if(aSize <= SETTING(PRIO_NORMAL_SIZE)*1024) {
			p = QueueItem::NORMAL;
		} else if(aSize <= SETTING(PRIO_LOW_SIZE)*1024) {
			p = QueueItem::LOW;
		} else if(BOOLSETTING(PRIO_LOWEST)) {
			p = QueueItem::LOWEST;
		} else {
			p = QueueItem::NORMAL;
		}
	}

	QueueItem* qi = new QueueItem(aTarget, aSize, p, aFlags, aDownloadedBytes, aAdded, root);

	if(!qi->isSet(QueueItem::FLAG_USER_LIST)) {
		if(!aTempTarget.empty()) {
			qi->setTempTarget(aTempTarget);
		}
	} else {
		qi->setPriority(QueueItem::HIGHEST);
	}

	if((qi->getDownloadedBytes() > 0))
		qi->setFlag(QueueItem::FLAG_EXISTS);

	dcassert(find(aTarget) == NULL);
	add(qi);
	return qi;
}

void QueueManager::FileQueue::add(QueueItem* qi) {
	if(lastInsert == queue.end())
		lastInsert = queue.insert(make_pair(const_cast<string*>(&qi->getTarget()), qi)).first;
	else
		lastInsert = queue.insert(lastInsert, make_pair(const_cast<string*>(&qi->getTarget()), qi));
}

QueueItem* QueueManager::FileQueue::find(const string& target) {
	QueueItem::StringIter i = queue.find(const_cast<string*>(&target));
	return (i == queue.end()) ? NULL : i->second;
}

void QueueManager::FileQueue::find(QueueItem::List& sl, int64_t aSize, const string& suffix) {
	for(QueueItem::StringIter i = queue.begin(); i != queue.end(); ++i) {
		if(i->second->getSize() == aSize) {
			const string& t = i->second->getTarget();
			if(suffix.empty() || (suffix.length() < t.length() &&
				Util::stricmp(suffix.c_str(), t.c_str() + (t.length() - suffix.length())) == 0) )
				sl.push_back(i->second);
		}
	}
}

void QueueManager::FileQueue::find(QueueItem::List& ql, const TTHValue& tth) {
	for(QueueItem::StringIter i = queue.begin(); i != queue.end(); ++i) {
		QueueItem* qi = i->second;
		if(qi->getTTH() != NULL && *qi->getTTH() == tth) {
			ql.push_back(qi);
		}
	}
}

static QueueItem* findCandidate(QueueItem::StringIter start, QueueItem::StringIter end, StringList& recent) {
	QueueItem* cand = NULL;
	for(QueueItem::StringIter i = start; i != end; ++i) {
		QueueItem* q = i->second;

		// We prefer to search for things that are not running...
		if((cand != NULL) && (q->getStatus() == QueueItem::STATUS_RUNNING))
			continue;
		// No user lists
		if(q->isSet(QueueItem::FLAG_USER_LIST))
			continue;
		// No paused downloads
		if(q->getPriority() == QueueItem::PAUSED)
			continue;
		// No files that already have more than AUTO_SEARCH_LIMIT online sources
		if(SETTING(AUTO_SEARCH_LIMIT) != 0 && q->countOnlineUsers() >= SETTING(AUTO_SEARCH_LIMIT))
			continue;
		// No files without TTH
		if(q->getTTH() == NULL)
			continue;
		// Did we search for it recently?
		if(find(recent.begin(), recent.end(), q->getTarget()) != recent.end())
			continue;

		cand = q;

		if(cand->getStatus() != QueueItem::STATUS_RUNNING)
			break;
	}
	return cand;
}

QueueItem* QueueManager::FileQueue::findAutoSearch(StringList& recent) {
	// We pick a start position at random, hoping that we will find something to search for...
	QueueItem::StringMap::size_type start = (QueueItem::StringMap::size_type)Util::rand((uint32_t)queue.size());

	QueueItem::StringIter i = queue.begin();
	advance(i, start);

	QueueItem* cand = findCandidate(i, queue.end(), recent);
	if(cand == NULL) {
		cand = findCandidate(queue.begin(), i, recent);
	} else if(cand->getStatus() == QueueItem::STATUS_RUNNING) {
		QueueItem* cand2 = findCandidate(queue.begin(), i, recent);
		if(cand2 != NULL && (cand2->getStatus() != QueueItem::STATUS_RUNNING)) {
			cand = cand2;
		}
	}
	return cand;
}

void QueueManager::FileQueue::move(QueueItem* qi, const string& aTarget) {
	if(lastInsert != queue.end() && Util::stricmp(*lastInsert->first, qi->getTarget()) == 0)
		lastInsert = queue.end();
	queue.erase(const_cast<string*>(&qi->getTarget()));
	qi->setTarget(aTarget);
	add(qi);
}

void QueueManager::UserQueue::add(QueueItem* qi) {
	for(QueueItem::SourceConstIter i = qi->getSources().begin(); i != qi->getSources().end(); ++i) {
		add(qi, i->getUser());
	}
}

void QueueManager::UserQueue::add(QueueItem* qi, const UserPtr& aUser) {
	if(qi->getStatus() == QueueItem::STATUS_RUNNING) {
		return;
	}

	QueueItem::List& l = userQueue[qi->getPriority()][aUser];
	if(qi->isSet(QueueItem::FLAG_EXISTS)) {
		l.insert(l.begin(), qi);
	} else {
		l.push_back(qi);
	}
}

QueueItem* QueueManager::UserQueue::getNext(const UserPtr& aUser, QueueItem::Priority minPrio) {
	int p = QueueItem::LAST - 1;

	do {
		QueueItem::UserListIter i = userQueue[p].find(aUser);
		if(i != userQueue[p].end()) {
			dcassert(!i->second.empty());
			return i->second.front();
		}
		p--;
	} while(p >= minPrio);

	return NULL;
}

void QueueManager::UserQueue::setRunning(QueueItem* qi, const UserPtr& aUser) {
	dcassert(qi->getCurrentDownload() == 0);
	dcassert(qi->getStatus() == QueueItem::STATUS_WAITING);

	// Remove the download from the userQueue...
	remove(qi);

	// Set the flag to running...
	qi->setStatus(QueueItem::STATUS_RUNNING);
	qi->setCurrent(aUser);

	// Move the download to the running list...
	dcassert(running.find(aUser) == running.end());
	running[aUser] = qi;
}

void QueueManager::UserQueue::setWaiting(QueueItem* qi) {
	dcassert(qi->getStatus() == QueueItem::STATUS_RUNNING);
	dcassert(qi->getCurrentDownload() != 0);

	dcassert(running.find(qi->getCurrent()) != running.end());
	// Remove the download from running
	running.erase(qi->getCurrent());

	// Set flag to waiting
	qi->setStatus(QueueItem::STATUS_WAITING);
	qi->setCurrent(UserPtr());
	qi->setCurrentDownload(0);

	// Add to the userQueue
	add(qi);
}

QueueItem* QueueManager::UserQueue::getRunning(const UserPtr& aUser) {
	QueueItem::UserIter i = running.find(aUser);
	return (i == running.end()) ? 0 : i->second;
}

QueueItem* QueueManager::getRunning(const UserPtr& aUser) {
	return userQueue.getRunning(aUser);
}

void QueueManager::UserQueue::remove(QueueItem* qi) {
	for(QueueItem::SourceConstIter i = qi->getSources().begin(); i != qi->getSources().end(); ++i) {
		remove(qi, i->getUser());
	}
}

void QueueManager::UserQueue::remove(QueueItem* qi, const UserPtr& aUser) {
	if(qi->getStatus() == QueueItem::STATUS_RUNNING) {
		if(qi->getCurrent() == aUser) {
			// Remove from running...
			dcassert(running.find(aUser) != running.end());
			running.erase(aUser);
		}
	} else {
		dcassert(qi->isSource(aUser));
		QueueItem::UserListMap& ulm = userQueue[qi->getPriority()];
		QueueItem::UserListIter j = ulm.find(aUser);
		dcassert(j != ulm.end());
		QueueItem::List& l = j->second;
		dcassert(find(l.begin(), l.end(), qi) != l.end());
		l.erase(std::remove(l.begin(), l.end(), qi), l.end());

		if(l.empty()) {
			ulm.erase(j);
		}
	}
}

QueueManager::QueueManager() : lastSave(0), queueFile(Util::getConfigPath() + "Queue.xml"), dirty(true), nextSearch(0) {
	TimerManager::getInstance()->addListener(this);
	SearchManager::getInstance()->addListener(this);
	ClientManager::getInstance()->addListener(this);

	File::ensureDirectory(Util::getListPath());
}

QueueManager::~QueueManager() throw() {
	SearchManager::getInstance()->removeListener(this);
	TimerManager::getInstance()->removeListener(this);
	ClientManager::getInstance()->removeListener(this);

	saveQueue();

	if(!BOOLSETTING(KEEP_LISTS)) {
		string path = Util::getListPath();
		StringList filelists = File::findFiles(path, "*.xml.bz2");
		StringList filelists2 = File::findFiles(path, "*.DcLst");
		filelists.insert(filelists.end(), filelists2.begin(), filelists2.end());

		for(StringIter i = filelists.begin(); i != filelists.end(); ++i) {
			File::deleteFile(*i);
		}
	}
}

bool QueueManager::getTTH(const string& name, TTHValue& tth) throw() {
	Lock l(cs);
	QueueItem* qi = fileQueue.find(name);
	if(qi != NULL && qi->getTTH() != NULL) {
		tth = *qi->getTTH();
		return true;
	}
	return false;
}

void QueueManager::on(TimerManagerListener::Minute, uint64_t aTick) throw() {
	string searchString;
	{
		Lock l(cs);
		QueueItem::UserMap& um = userQueue.getRunning();
		if(!um.empty())
			setDirty();

		if(BOOLSETTING(AUTO_SEARCH) && (!BOOLSETTING(AUTO_SEARCH_ON_AWAY) || Util::getAway()) && (aTick >= nextSearch) && (fileQueue.getSize() > 0) && SearchManager::getInstance()->okToSearch()) {
			// We keep 32 recent searches to avoid duplicate searches
			while((recent.size() >= fileQueue.getSize()) || (recent.size() > 32)) {
				recent.erase(recent.begin());
			}

			QueueItem* qi = fileQueue.findAutoSearch(recent);
			if(qi == NULL && (recent.size() > 0 && fileQueue.getSize() > 0)) {
				recent.erase(recent.begin());
				qi = fileQueue.findAutoSearch(recent);
            }
			if(qi != NULL) {
				dcassert(qi->getTTH());
				searchString = qi->getTTH()->toBase32();
				recent.push_back(qi->getTarget());
				nextSearch = aTick + (fileQueue.getSize() > 10 ? 300000 : 600000);
				if(BOOLSETTING(REPORT_ALTERNATES)) {
					LogManager::getInstance()->message(CSTRING(ALTERNATES_SEND) + Util::getFileName(qi->getTargetFileName()), true);
				}
			}
		}
	}

	if(!searchString.empty()) {
		SearchManager::getInstance()->search(searchString, 0, SearchManager::TYPE_TTH, SearchManager::SIZE_DONTCARE);
	}

	if(dirty && ((lastSave + 120000) < aTick)) {
		saveQueue();
	}
}

void QueueManager::addList(const UserPtr& aUser, Flags<uint16_t>::MaskType aFlags) throw(QueueException, FileException) {
    string target = Util::getListPath() + Util::cleanPathChars(aUser->getFirstNick()) + "." + aUser->getCID().toBase32();
	add(target, -1, NULL, aUser, USER_LIST_NAME, true, (Flags<uint16_t>::MaskType)(QueueItem::FLAG_USER_LIST | aFlags));
}

void QueueManager::add(const string& aTarget, int64_t aSize, const TTHValue* root, UserPtr aUser, const string& aSourceFile,
					   bool utf8, Flags<uint16_t>::MaskType aFlags /* = QueueItem::FLAG_RESUME */, bool addBad /* = true */) throw(QueueException, FileException) {
	bool wantConnection = true;

	// Check that we're not downloading from ourselves...
	if(aUser == ClientManager::getInstance()->getMe()) {
		throw QueueException(STRING(NO_DOWNLOADS_FROM_SELF));
	}

	// Check if we're not downloading something already in our share
	if(BOOLSETTING(DONT_DL_ALREADY_SHARED) && root != NULL) {
		if(ShareManager::getInstance()->isTTHShared(*root)) {
			throw QueueException(STRING(TTH_ALREADY_SHARED));
		}
	}

	string target = checkTarget(aTarget, aSize, aFlags);

	// Check if it's a zero-byte file, if so, create and return...
	if(aSize == 0) {
		if(!BOOLSETTING(SKIP_ZERO_BYTE)) {
			File::ensureDirectory(target);
			File f(target, File::WRITE, File::CREATE);
		}
		return;
	}

	// Check if we're trying to download a non-TTH file
	if(root == NULL && !(aFlags &QueueItem::FLAG_USER_LIST)) {
		if(BOOLSETTING(ONLY_DL_TTH_FILES))
			throw QueueException(STRING(FILE_HAS_NO_TTH));
	}

	if(aUser->isSet(User::PASSIVE) && !ClientManager::getInstance()->isActive()) {
		throw QueueException(STRING(NO_DOWNLOADS_FROM_PASSIVE));
	}

	{
		Lock l(cs);

		QueueItem* q = fileQueue.find(target);
		if(q == NULL) {
			if(BOOLSETTING(DONT_DL_ALREADY_QUEUED) && root != NULL) {
				QueueItem::List matches;

				fileQueue.find(matches, *root);

				for(QueueItem::Iter i = matches.begin(); i != matches.end(); ++i) {
					QueueItem* qi = *i;

					if(*qi->getTTH() == *root && aSize == qi->getSize()) {
						throw QueueException(STRING(FILE_IS_ALREADY_QUEUED));
					}
				}
			}
			q = fileQueue.add(target, aSize, aFlags, QueueItem::DEFAULT, Util::emptyString, 0, GET_TIME(), root);
			fire(QueueManagerListener::Added(), q);
		} else {
			if(q->getSize() != aSize) {
				throw QueueException(STRING(FILE_WITH_DIFFERENT_SIZE));
			}
			if(root != NULL) {
				if(q->getTTH() == NULL) {
					q->setTTH(new TTHValue(*root));
				} else if(*root != *q->getTTH()) {
					throw QueueException(STRING(FILE_WITH_DIFFERENT_TTH));
				}
			}
			q->setFlag(aFlags);

			// We don't add any more sources to user list downloads...
			if(q->isSet(QueueItem::FLAG_USER_LIST))
				return;
		}

		wantConnection = addSource(q, aSourceFile, aUser, (Flags<uint16_t>::MaskType)(addBad ? QueueItem::Source::FLAG_MASK : 0), utf8);
	}

	if(wantConnection && aUser->isOnline())
		ConnectionManager::getInstance()->getDownloadConnection(aUser);
}

void QueueManager::readd(const string& target, UserPtr& aUser) throw(QueueException) {
	bool wantConnection = false;
	{
		Lock l(cs);
		QueueItem* q = fileQueue.find(target);
		if(q != NULL && q->isBadSource(aUser)) {
			QueueItem::SourceIter s = q->getBadSource(aUser);
			wantConnection = addSource(q, s->getPath(), aUser, QueueItem::Source::FLAG_MASK, s->isSet(QueueItem::Source::FLAG_UTF8));
		}
	}

	if(wantConnection && aUser->isOnline())
		ConnectionManager::getInstance()->getDownloadConnection(aUser);
}

void QueueManager::readdall(const string& target) throw(QueueException) {
	Lock l(cs);
	QueueItem* q = fileQueue.find(target);
	if(q != NULL) {
		QueueItem::SourceIter j = q->getBadSources().begin();
		while(j != q->getBadSources().end()) {
			UserPtr user = j->getUser();
			bool wantConnection = addSource(q, j->getPath(), user, QueueItem::Source::FLAG_MASK, j->isSet(QueueItem::Source::FLAG_UTF8));
			if(wantConnection && user->isOnline())
				ConnectionManager::getInstance()->getDownloadConnection(user);
		}
	}
}

string QueueManager::checkTarget(const string& aTarget, int64_t aSize, Flags<uint16_t>::MaskType& flags) throw(QueueException, FileException) {
#ifdef _WIN32
	if(aTarget.length() > MAX_PATH) {
		throw QueueException(STRING(TARGET_FILENAME_TOO_LONG));
	}
	// Check that target starts with a drive or is an UNC path
	if( (aTarget[1] != ':' || aTarget[2] != '\\') &&
		(aTarget[0] != '\\' && aTarget[1] != '\\') ) {
		throw QueueException(STRING(INVALID_TARGET_FILE));
	}
#else
	if(aTarget.length() > PATH_MAX) {
		throw QueueException(STRING(TARGET_FILENAME_TOO_LONG));
	}
	// Check that target contains at least one directory...we don't want headless files...
	if(aTarget[0] != '/') {
		throw QueueException(STRING(INVALID_TARGET_FILE));
	}
#endif

	string target = Util::validateFileName(aTarget);

	// Check that the file doesn't already exist...
	int64_t sz = File::getSize(target);
	if( (aSize != -1) && (aSize <= sz) ) {
		throw FileException(STRING(LARGER_TARGET_FILE_EXISTS));
	}
	if(sz > 0)
		flags |= QueueItem::FLAG_EXISTS;

	return target;
}

/** Add a source to an existing queue item */
bool QueueManager::addSource(QueueItem* qi, const string& aFile, UserPtr aUser, Flags<uint16_t>::MaskType addBad, bool utf8) throw(QueueException, FileException) {
	bool wantConnection = (qi->getPriority() != QueueItem::PAUSED) && (qi->getStatus() != QueueItem::STATUS_RUNNING);

	if(qi->isSource(aUser)) {
		throw QueueException(STRING(DUPLICATE_SOURCE) + ": " + Util::getFileName(qi->getTarget()) + ", " + aUser->getFirstNick());
	}

	if(qi->isBadSourceExcept(aUser, addBad)) {
		throw QueueException(STRING(DUPLICATE_SOURCE) + ": " + Util::getFileName(qi->getTarget()) + ", " + aUser->getFirstNick());
	}

	qi->addSource(aUser, aFile);

	if(utf8) {
		QueueItem::SourceIter s = qi->getSource(aUser);
		s->setFlag(QueueItem::Source::FLAG_UTF8);
	}

	if(aUser->isSet(User::PASSIVE) && !ClientManager::getInstance()->isActive() ) {
		qi->removeSource(aUser, QueueItem::Source::FLAG_PASSIVE);
		wantConnection = false;
	} else {
		userQueue.add(qi, aUser);
		aUser->setFlag(User::SAVE_NICK);
	}

	fire(QueueManagerListener::SourcesUpdated(), qi);
	setDirty();

	return wantConnection;
}

void QueueManager::addDirectory(const string& aDir, const UserPtr& aUser, const string& aTarget, QueueItem::Priority p /* = QueueItem::DEFAULT */) throw() {
	bool needList;
	{
		Lock l(cs);

		DirectoryItem::DirectoryPair dp = directories.equal_range(aUser);

		for(DirectoryItem::DirectoryIter i = dp.first; i != dp.second; ++i) {
			if(Util::stricmp(aTarget.c_str(), i->second->getName().c_str()) == 0)
				return;
		}

		// Unique directory, fine...
		directories.insert(make_pair(aUser, new DirectoryItem(aUser, aDir, aTarget, p)));
		needList = (dp.first == dp.second);
		setDirty();
	}

	if(needList) {
		try {
			addList(aUser, QueueItem::FLAG_DIRECTORY_DOWNLOAD);
		} catch(const Exception&) {
			// Ignore, we don't really care...
		}
	}
}

QueueItem::Priority QueueManager::hasDownload(const UserPtr& aUser) throw() {
	Lock l(cs);
	QueueItem* qi = userQueue.getNext(aUser, QueueItem::LOWEST);
	if(qi == NULL) {
		return QueueItem::PAUSED;
	}
	return qi->getPriority();
}

namespace {
	typedef HASH_MAP_X(TTHValue*, const DirectoryListing::File*, TTHValue::PtrHash, TTHValue::PtrHash, TTHValue::PtrLess) TTHMap;

	// *** WARNING ***
	// Lock(cs) makes sure that there's only one thread accessing this
	static TTHMap tthMap;

	void buildMap(const DirectoryListing::Directory* dir) throw() {
		for(DirectoryListing::Directory::List::const_iterator j = dir->directories.begin(); j != dir->directories.end(); ++j) {
			if((*j)->getAdls() == false)
				buildMap(*j);
		}

		for(DirectoryListing::File::List::const_iterator i = dir->files.begin(); i != dir->files.end(); ++i) {
			const DirectoryListing::File* df = *i;
			if(df->getTTH() != NULL)
                tthMap.insert(make_pair(df->getTTH(), df));
		}
	}
}

int QueueManager::matchListing(const DirectoryListing& dl) throw() {
	int matches = 0;
	{
		Lock l(cs);
		tthMap.clear();
		buildMap(dl.getRoot());

        if(tthMap.size() > 0) {
    		for(QueueItem::StringMap::const_iterator i = fileQueue.getQueue().begin(); i != fileQueue.getQueue().end(); ++i) {
    			QueueItem* qi = i->second;

    			if(qi->isSet(QueueItem::FLAG_USER_LIST))
    				continue;

    			if(qi->getTTH() == NULL)
    				continue;

    			TTHMap::iterator j = tthMap.find(qi->getTTH());
    			if(j != tthMap.end() && j->second->getSize() == qi->getSize()) {
    				try {
    					addSource(qi, dl.getPath(j->second) + j->second->getName(), dl.getUser(), QueueItem::Source::FLAG_FILE_NOT_AVAILABLE, dl.getUtf8());
    					matches++;
    				} catch(...) {
    					// Ignore...
    				}
    			}
    		}
        }
        tthMap.clear();
	}
	if(matches > 0)
		ConnectionManager::getInstance()->getDownloadConnection(dl.getUser());

	return matches;
}

void QueueManager::move(const string& aSource, const string& aTarget) throw() {
	string target = Util::validateFileName(aTarget);
	if(aSource == target)
		return;

	bool delSource = false;

	Lock l(cs);
	QueueItem* qs = fileQueue.find(aSource);
	if(qs != NULL) {
		// Don't move running downloads
		if(qs->getStatus() == QueueItem::STATUS_RUNNING) {
			return;
		}
		// Don't move file lists
		if(qs->isSet(QueueItem::FLAG_USER_LIST))
			return;

		// Let's see if the target exists...then things get complicated...
		QueueItem* qt = fileQueue.find(target);
		if(qt == NULL || Util::stricmp(aSource, target) == 0) {
			// Good, update the target and move in the queue...
			fileQueue.move(qs, target);
			fire(QueueManagerListener::Moved(), qs, aSource);
			setDirty();
		} else {
			// Don't move to target of different size
			if(qs->getSize() != qt->getSize())
				return;

			try {
				for(QueueItem::SourceConstIter i = qs->getSources().begin(); i != qs->getSources().end(); ++i) {
					addSource(qt, i->getPath(), i->getUser(), QueueItem::Source::FLAG_MASK, i->isSet(QueueItem::Source::FLAG_UTF8));
				}
			} catch(const Exception&) {
			}
			delSource = true;
		}
	}

	if(delSource) {
		remove(aSource);
	}
}

bool QueueManager::getQueueInfo(UserPtr& aUser, string& aTarget, int64_t& aSize) throw() {
    Lock l(cs);
    QueueItem* qi = userQueue.getNext(aUser);
	if(qi == NULL)
		return false;

	aTarget = qi->getTarget();
	aSize = qi->getSize();

	return true;
}

void QueueManager::getTargetsBySize(StringList& sl, int64_t aSize, const string& suffix) throw() {
	Lock l(cs);
	QueueItem::List ql;
	fileQueue.find(ql, aSize, suffix);
	for(QueueItem::Iter i = ql.begin(); i != ql.end(); ++i) {
		sl.push_back((*i)->getTarget());
	}
}

void QueueManager::getTargetsByRoot(StringList& sl, const TTHValue& tth) {
	Lock l(cs);
	QueueItem::List ql;
	fileQueue.find(ql, tth);
	for(QueueItem::Iter i = ql.begin(); i != ql.end(); ++i) {
		sl.push_back((*i)->getTarget());
	}
}

Download* QueueManager::getDownload(UserConnection& aSource, bool supportsTrees) throw() {
	Lock l(cs);

	UserPtr& aUser = aSource.getUser();
	QueueItem* q = userQueue.getNext(aUser);

	if(q == NULL)
		return NULL;

	userQueue.setRunning(q, aUser);

	Download* d = new Download(aSource, *q);

	q->setCurrentDownload(d);

	QueueItem::SourceIter src = q->getSource(aUser);

	if(aSource.isSet(UserConnection::FLAG_SUPPORTS_ADCGET) && aSource.isSet(UserConnection::FLAG_SUPPORTS_TTHF) && q->getTTH() != NULL) {
		src->setPath(Util::emptyString);
		d->setSource(Util::emptyString);
	} else {
		d->setSource(src->getPath());
		if(src->isSet(QueueItem::Source::FLAG_UTF8))
			d->setFlag(Download::FLAG_UTF8);
	}

	if(d->getSize() != -1 && d->getTTH() != NULL) {
		if(HashManager::getInstance()->getTree(*d->getTTH(), d->getTigerTree())) {
			d->setStartPos(q->getDownloadedBytes());
			d->setTreeValid(true);
		} else if(supportsTrees && !q->getSource(aUser)->isSet(QueueItem::Source::FLAG_NO_TREE) && d->getSize() > HashManager::MIN_BLOCK_SIZE) {
			// Get the tree unless the file is small (for small files, we'd probably only get the root anyway)
			d->setFlag(Download::FLAG_TREE_DOWNLOAD);
			d->getTigerTree().setFileSize(d->getSize());
			d->setPos(0);
			d->setSize(-1);
			d->unsetFlag(Download::FLAG_RESUME);

            if(aSource.isSet(UserConnection::FLAG_SUPPORTS_TTHF) == false) {
                d->setSource(Text::toUtf8(src->getPath()));
                d->setFlag(Download::FLAG_UTF8);
            }
		} else {
			// Use the root as tree to get some sort of validation at least...
			d->getTigerTree() = TigerTree(d->getSize(), d->getSize(), *d->getTTH());
			d->setStartPos(q->getDownloadedBytes());
			d->setTreeValid(true);
		}
	}

	if(!d->isSet(Download::FLAG_TREE_DOWNLOAD) && BOOLSETTING(ANTI_FRAG) ) {
		d->setStartPos(q->getDownloadedBytes());
	}

	fire(QueueManagerListener::StatusUpdated(), q);
	return d;
}

void QueueManager::putDownload(Download* aDownload, bool finished) throw() {
	UserList getConn;
	string fname;
	UserPtr up;
	Flags<uint16_t>::MaskType flag = 0;

	{
		Lock l(cs);
		QueueItem* q = fileQueue.find(aDownload->getTarget());

		if(q != NULL) {
			if(aDownload->isSet(Download::FLAG_USER_LIST)) {
				if(aDownload->getSource() == Transfer::USER_LIST_NAME_BZ) {
					q->setFlag(QueueItem::FLAG_XML_BZLIST);
			} else {
				q->unsetFlag(QueueItem::FLAG_XML_BZLIST);
				}
			}

			if(finished) {
				if(aDownload->isSet(Download::FLAG_TREE_DOWNLOAD)) {
					// Got a full tree, now add it to the HashManager
					dcassert(aDownload->getTreeValid());
					HashManager::getInstance()->addTree(aDownload->getTigerTree());

					if(q->getStatus() == QueueItem::STATUS_RUNNING) {
						userQueue.setWaiting(q);
						fire(QueueManagerListener::StatusUpdated(), q);
					}
				} else {
					// Now, let's see if this was a directory download filelist...
					if( (q->isSet(QueueItem::FLAG_DIRECTORY_DOWNLOAD) && directories.find(q->getCurrent()) != directories.end()) ||
						(q->isSet(QueueItem::FLAG_MATCH_QUEUE)) ) 
					{
						fname = q->getListName();
						up = q->getCurrent();
						flag = (Flags<uint16_t>::MaskType)((q->isSet(QueueItem::FLAG_DIRECTORY_DOWNLOAD) ? QueueItem::FLAG_DIRECTORY_DOWNLOAD : 0)
							| (q->isSet(QueueItem::FLAG_MATCH_QUEUE) ? QueueItem::FLAG_MATCH_QUEUE : 0));
					} 

					fire(QueueManagerListener::Finished(), q, aDownload->getAverageSpeed());
					fire(QueueManagerListener::Removed(), q);

					userQueue.remove(q);
					fileQueue.remove(q);
					setDirty();
				}
			} else {
				if(!aDownload->isSet(Download::FLAG_TREE_DOWNLOAD)) {
					q->setDownloadedBytes(aDownload->getPos());

					if(q->getDownloadedBytes() > 0) {
						q->setFlag(QueueItem::FLAG_EXISTS);
					} else {
						q->setTempTarget(Util::emptyString);
					}
					if(q->isSet(QueueItem::FLAG_USER_LIST)) {
						// Blah...no use keeping an unfinished file list...
						File::deleteFile(q->getListName());
					}
				}

				if(q->getPriority() != QueueItem::PAUSED) {
						q->getOnlineUsers(getConn);
					}

					// This might have been set to wait elsewhere already...
				if(q->getStatus() == QueueItem::STATUS_RUNNING) {
					userQueue.setWaiting(q);
					fire(QueueManagerListener::StatusUpdated(), q);
				}
			}
		} else if(!aDownload->isSet(Download::FLAG_TREE_DOWNLOAD)) {
			if(!aDownload->getTempTarget().empty() && (aDownload->isSet(Download::FLAG_USER_LIST) || aDownload->getTempTarget() != aDownload->getTarget())) {
				File::deleteFile(aDownload->getTempTarget() + Download::ANTI_FRAG_EXT);
				File::deleteFile(aDownload->getTempTarget());
			}
		}
		delete aDownload;
	}

	for(UserList::iterator i = getConn.begin(); i != getConn.end(); ++i) {
		ConnectionManager::getInstance()->getDownloadConnection(*i);
	}

	if(!fname.empty()) {
		processList(fname, up, flag);
	}
}

void QueueManager::processList(const string& name, UserPtr& user, Flags<uint16_t>::MaskType flags) {
	DirectoryListing dirList(user);
	try {
		dirList.loadFile(name);
	} catch(const Exception&) {
		LogManager::getInstance()->message(STRING(UNABLE_TO_OPEN_FILELIST) + name);
		return;
	}

	if(flags & QueueItem::FLAG_DIRECTORY_DOWNLOAD) {
		DirectoryItem::List dl;
		{
			Lock l(cs);
			DirectoryItem::DirectoryPair dp = directories.equal_range(user);
			for(DirectoryItem::DirectoryIter i = dp.first; i != dp.second; ++i) {
				dl.push_back(i->second);
			}
			directories.erase(user);
		}

		for(DirectoryItem::Iter i = dl.begin(); i != dl.end(); ++i) {
			DirectoryItem* di = *i;
			dirList.download(di->getName(), di->getTarget(), false);
			delete di;
		}
	}
	if(flags & QueueItem::FLAG_MATCH_QUEUE) {
		const size_t BUF_SIZE = STRING(MATCHED_FILES).size() + 16;
		AutoArray<char> tmp(BUF_SIZE);
		sprintf_s(tmp, BUF_SIZE, CSTRING(MATCHED_FILES), matchListing(dirList));
		LogManager::getInstance()->message(Util::toString(ClientManager::getInstance()->getNicks(user->getCID())) + ": " + string(tmp));						
	}
}

void QueueManager::remove(const string& aTarget) throw() {
	UserPtr x;

	{
		Lock l(cs);

		QueueItem* q = fileQueue.find(aTarget);
		if(!q)
			return;

		if(q->isSet(QueueItem::FLAG_DIRECTORY_DOWNLOAD)) {
			dcassert(q->getSources().size() == 1);
			DirectoryItem::DirectoryPair dp = directories.equal_range(q->getSources()[0].getUser());
			for(DirectoryItem::DirectoryIter i = dp.first; i != dp.second; ++i) {
				delete i->second;
			}
			directories.erase(q->getSources()[0].getUser());
		}

		if(q->getStatus() == QueueItem::STATUS_RUNNING) {
			x = q->getCurrent();
		} else if(!q->getTempTarget().empty() && q->getTempTarget() != q->getTarget()) {
			File::deleteFile(q->getTempTarget() + Download::ANTI_FRAG_EXT);
			File::deleteFile(q->getTempTarget());
		}

		fire(QueueManagerListener::Removed(), q);

		userQueue.remove(q);
		fileQueue.remove(q);

		setDirty();
	}

	if(x) {
		ConnectionManager::getInstance()->disconnect(x, true);
	}
}

void QueueManager::removeSource(const string& aTarget, UserPtr& aUser, Flags<uint16_t>::MaskType reason, bool removeConn /* = true */) throw() {
	bool isRunning = false;
	bool removeCompletely = false;
	{
		Lock l(cs);
		QueueItem* q = fileQueue.find(aTarget);
		if(!q)
			return;

		if(!q->isSource(aUser))
			return;

		if(q->isSet(QueueItem::FLAG_USER_LIST)) {
			removeCompletely = true;
			goto endCheck;
		}

		if(reason == QueueItem::Source::FLAG_NO_TREE) {
			q->getSource(aUser)->setFlag(reason);
			return;
		}

		if(reason == QueueItem::Source::FLAG_CRC_WARN) {
			// Already flagged?
			QueueItem::SourceIter s = q->getSource(aUser);
			if(s->isSet(QueueItem::Source::FLAG_CRC_WARN)) {
				reason = QueueItem::Source::FLAG_CRC_FAILED;
			} else {
				s->setFlag(reason);
				return;
			}
		}

		if((q->getStatus() == QueueItem::STATUS_RUNNING) && q->getCurrent() == aUser) {
			isRunning = true;
			userQueue.setWaiting(q);
			fire(QueueManagerListener::StatusUpdated(), q);
		}

		userQueue.remove(q, aUser);
		q->removeSource(aUser, reason);

		fire(QueueManagerListener::SourcesUpdated(), q);
		setDirty();
	}
endCheck:
	if(isRunning && removeConn) {
		ConnectionManager::getInstance()->disconnect(aUser, true);
	}
	if(removeCompletely) {
		remove(aTarget);
	}
}

void QueueManager::removeAllSources(const string& aTarget) throw() {
	Lock l(cs);
	QueueItem* q = fileQueue.find(aTarget);
	if(q == NULL)
		return;

	if((q->getStatus() == QueueItem::STATUS_RUNNING)) {
		UserPtr user = q->getCurrent();
		userQueue.setWaiting(q);
		fire(QueueManagerListener::StatusUpdated(), q);
		ConnectionManager::getInstance()->disconnect(user, true);
	}

	QueueItem::SourceIter j = q->getSources().begin();
	while(j != q->getSources().end()) {
		UserPtr user = j->getUser();
		userQueue.remove(q, user);
		q->removeSource(user, QueueItem::Source::FLAG_REMOVED);
	}

	fire(QueueManagerListener::SourcesUpdated(), q);
	setDirty();
}

void QueueManager::removeOffline(const string& aTarget) throw() {
	Lock l(cs);
	QueueItem* q = fileQueue.find(aTarget);
	if(q == NULL)
		return;

	QueueItem::SourceIter j = q->getSources().begin();
	while(j != q->getSources().end()) {
		UserPtr user = j->getUser();
		if(user->isOnline() == false) {
            userQueue.remove(q, user);
            q->removeSource(user, QueueItem::Source::FLAG_REMOVED);
        } else {
            j++;
        }
	}

	fire(QueueManagerListener::SourcesUpdated(), q);
	setDirty();
}

void QueueManager::removeSource(UserPtr& aUser, Flags<uint16_t>::MaskType reason) throw() {
	bool isRunning = false;
	string removeRunning;
	{
		Lock l(cs);
		QueueItem* qi = NULL;
		while( (qi = userQueue.getNext(aUser, QueueItem::PAUSED)) != NULL) {
			if(qi->isSet(QueueItem::FLAG_USER_LIST)) {
				remove(qi->getTarget());
			} else {
				userQueue.remove(qi, aUser);
				qi->removeSource(aUser, reason);
				fire(QueueManagerListener::SourcesUpdated(), qi);
				setDirty();
			}
		}

		qi = userQueue.getRunning(aUser);
		if(qi) {
			if(qi->isSet(QueueItem::FLAG_USER_LIST)) {
				removeRunning = qi->getTarget();
			} else {
				userQueue.setWaiting(qi);
				userQueue.remove(qi, aUser);
				isRunning = true;
				qi->removeSource(aUser, reason);
				fire(QueueManagerListener::StatusUpdated(), qi);
				fire(QueueManagerListener::SourcesUpdated(), qi);
				setDirty();
			}
		}
	}

	if(isRunning) {
		ConnectionManager::getInstance()->disconnect(aUser, true);
	}
	if(!removeRunning.empty()) {
		remove(removeRunning);
	}
}

void QueueManager::setPriority(const string& aTarget, QueueItem::Priority p) throw() {
	UserList ul;
	bool running = false;

	{
		Lock l(cs);

		QueueItem* q = fileQueue.find(aTarget);
		if( (q != NULL) && (q->getPriority() != p) ) {
			if( q->getStatus() == QueueItem::STATUS_WAITING ) {
				if(q->getPriority() == QueueItem::PAUSED || p == QueueItem::HIGHEST) {
					// Problem, we have to request connections to all these users...
					q->getOnlineUsers(ul);
				}

				userQueue.remove(q);
				q->setPriority(p);
				userQueue.add(q);
			} else {
				running = true;
				q->setPriority(p);
			}
			setDirty();
			fire(QueueManagerListener::StatusUpdated(), q);
		}
	}

	if(p == QueueItem::PAUSED) {
		if(running) {
			DownloadManager::getInstance()->abortDownload(aTarget);
		}
	} else {
		for(UserList::iterator i = ul.begin(); i != ul.end(); ++i) {
			ConnectionManager::getInstance()->getDownloadConnection(*i);
		}
    }
}

void QueueManager::saveQueue() throw() {
	if(!dirty)
		return;

	Lock l(cs);

	try {

		File ff(getQueueFile() + ".tmp", File::WRITE, File::CREATE | File::TRUNCATE);
		BufferedOutputStream<false> f(&ff);

		f.write(SimpleXML::utf8Header);
		f.write(LIT("<Downloads Version=\"" VERSIONSTRING "\">\r\n"));
		string tmp;
		string b32tmp;
		for(QueueItem::StringIter i = fileQueue.getQueue().begin(); i != fileQueue.getQueue().end(); ++i) {
			QueueItem* qi = i->second;
			if(!qi->isSet(QueueItem::FLAG_USER_LIST)) {
				f.write(LIT("\t<Download Target=\""));
				f.write(SimpleXML::escape(qi->getTarget(), tmp, true));
				f.write(LIT("\" Size=\""));
				f.write(Util::toString(qi->getSize()));
				f.write(LIT("\" Priority=\""));
				f.write(Util::toString((int)qi->getPriority()));
				f.write(LIT("\" Added=\""));
				f.write(Util::toString(qi->getAdded()));
				if(qi->getTTH() != NULL) {
					b32tmp.clear();
					f.write(LIT("\" TTH=\""));
					f.write(qi->getTTH()->toBase32(b32tmp));
				}
				if(qi->getDownloadedBytes() > 0) {
					f.write(LIT("\" TempTarget=\""));
					f.write(SimpleXML::escape(qi->getTempTarget(), tmp, true));
					f.write(LIT("\" Downloaded=\""));
					f.write(Util::toString(qi->getDownloadedBytes()));
				}
				f.write(LIT("\">\r\n"));

				for(QueueItem::SourceConstIter j = qi->sources.begin(); j != qi->sources.end(); ++j) {
					f.write(LIT("\t\t<Source CID=\""));
					f.write(j->getUser()->getCID().toBase32());

					if(!j->getPath().empty() && (!j->getUser()->isSet(User::TTH_GET) || !qi->getTTH()) ) {
						f.write(LIT("\" Path=\""));
						f.write(SimpleXML::escape(j->getPath(), tmp, true));
						f.write(LIT("\" Utf8=\""));
    					f.write(j->isSet(QueueItem::Source::FLAG_UTF8) ? "1" : "0", 1);
					}
					f.write(LIT("\"/>\r\n"));
				}

				f.write(LIT("\t</Download>\r\n"));
			}
		}

		f.write("</Downloads>\r\n");
		f.flush();
		ff.close();

		File::deleteFile(getQueueFile() + ".bak");
		CopyFile(Text::toT(getQueueFile()).c_str(), Text::toT(getQueueFile() + ".bak").c_str(), FALSE);
		File::deleteFile(getQueueFile());
		File::renameFile(getQueueFile() + ".tmp", getQueueFile());

		ClientManager::getInstance()->save();

		dirty = false;
	} catch(const FileException&) {
		// ...
	}
	// Put this here to avoid very many saves tries when disk is full...
	lastSave = GET_TICK();
}

class QueueLoader : private SimpleXMLReader::CallBack {
public:
	QueueLoader() : cur(NULL), inDownloads(false) { }
	~QueueLoader() { }
	void startTag(const string& name, StringPairList& attribs, bool simple);
	void endTag(const string& name, const string& data);
private:
	string target;

	QueueItem* cur;
	bool inDownloads;
};

void QueueManager::loadQueue() throw() {
	try {
		QueueLoader l;
		SimpleXMLReader(&l).fromXML(File(getQueueFile(), File::READ, File::OPEN).read());
		dirty = false;
	} catch(const Exception&) {
		// ...
	}
}

int QueueManager::countOnlineSources(const string& aTarget) {
	Lock l(cs);

	QueueItem* qi = fileQueue.find(aTarget);
	if(qi == NULL)
		return 0;

	int onlineSources = 0;
	for(QueueItem::SourceConstIter i = qi->getSources().begin(); i != qi->getSources().end(); ++i) {
		if(i->getUser()->isOnline())
			onlineSources++;
	}
	return onlineSources;
}

static const string sDownload = "Download";
static const string sTempTarget = "TempTarget";
static const string sTarget = "Target";
static const string sSize = "Size";
static const string sDownloaded = "Downloaded";
static const string sPriority = "Priority";
static const string sSource = "Source";
static const string sNick = "Nick";
static const string sPath = "Path";
static const string sDirectory = "Directory";
static const string sAdded = "Added";
static const string sUtf8 = "Utf8";
static const string sTTH = "TTH";
static const string sCID = "CID";

void QueueLoader::startTag(const string& name, StringPairList& attribs, bool simple) {
	QueueManager* qm = QueueManager::getInstance();
	if(!inDownloads && name == "Downloads") {
		inDownloads = true;
	} else if(inDownloads) {
		if(cur == NULL && name == sDownload) {
			Flags<uint16_t>::MaskType flags = QueueItem::FLAG_RESUME;
			int64_t size = Util::toInt64(getAttrib(attribs, sSize, 1));
			if(size == 0)
				return;
			try {
				const string& tgt = getAttrib(attribs, sTarget, 0);
				target = QueueManager::checkTarget(tgt, size, flags);
				if(target.empty())
					return;
			} catch(const Exception&) {
				return;
			}
			QueueItem::Priority p = (QueueItem::Priority)Util::toInt(getAttrib(attribs, sPriority, 3));
			time_t added = (time_t)Util::toInt(getAttrib(attribs, sAdded, 4));
			const string& tthRoot = getAttrib(attribs, sTTH, 5);
			string tempTarget = getAttrib(attribs, sTempTarget, 5);
			int64_t downloaded = Util::toInt64(getAttrib(attribs, sDownloaded, 5));
			if (downloaded > size || downloaded < 0)
				downloaded = 0;

			if(added == 0)
				added = GET_TIME();

			QueueItem* qi = qm->fileQueue.find(target);

			if(qi == NULL) {
				if(tthRoot.empty()) {
					qi = qm->fileQueue.add(target, size, flags, p, tempTarget, downloaded, added, NULL);
				} else {
					TTHValue root(tthRoot);
					qi = qm->fileQueue.add(target, size, flags, p, tempTarget, downloaded, added, &root);
				}
				qm->fire(QueueManagerListener::Added(), qi);
			}
			if(!simple)
				cur = qi;
		} else if(cur != NULL && name == sSource) {
			const string& cid = getAttrib(attribs, sCID, 0);
			if(cid.length() != 39) {
				// Skip loading this source - sorry old users
				return;
			}
			const string& path = getAttrib(attribs, sPath, 1);
			const string& utf8 = getAttrib(attribs, sUtf8, 2);
			bool isUtf8 = (utf8 == "1");
			UserPtr user = ClientManager::getInstance()->getUser(CID(cid));

			try {
				if(qm->addSource(cur, path, user, 0, isUtf8) && user->isOnline())
					ConnectionManager::getInstance()->getDownloadConnection(user);
			} catch(const Exception&) {
				return;
			}
		}
	}
}

void QueueLoader::endTag(const string& name, const string&) {
	if(inDownloads) {
		if(name == sDownload)
			cur = NULL;
		else if(name == "Downloads")
			inDownloads = false;
	}
}

// SearchManagerListener
void QueueManager::on(SearchManagerListener::SR, SearchResult* sr) throw() {
	bool added = false;
	bool wantConnection = false;
	int users = 0;

	if(BOOLSETTING(AUTO_SEARCH) && sr->getTTH() != NULL) {
		Lock l(cs);
		QueueItem::List matches;

		fileQueue.find(matches, *sr->getTTH());

		for(QueueItem::Iter i = matches.begin(); i != matches.end(); ++i) {
			QueueItem* qi = *i;
			dcassert(qi->getTTH());

			// Size compare to avoid popular spoof
			bool found = (*qi->getTTH() == *sr->getTTH()) && (sr->getSize() == qi->getSize());

			if(found && !qi->isSource(sr->getUser())) {
				try {
					if(!BOOLSETTING(AUTO_SEARCH_AUTO_MATCH))
						wantConnection = addSource(qi, sr->getFile(), sr->getUser(), 0, false);
					added = true;
					users = qi->countOnlineUsers();
				} catch(const Exception&) {
					// ...
				}
				break;
			}
		}
	}

	if((added && BOOLSETTING(AUTO_SEARCH_AUTO_MATCH)) && ((users < SETTING(MAX_FILELIST2MATCH)) || (SETTING(MAX_FILELIST2MATCH) == 0))) {
		try {
			addList(sr->getUser(), QueueItem::FLAG_MATCH_QUEUE);
		} catch(const Exception&) {
			// ...
		}
	}
	if(added && wantConnection && sr->getUser()->isOnline()) {
		ConnectionManager::getInstance()->getDownloadConnection(sr->getUser());
	}
}

// ClientManagerListener
void QueueManager::on(ClientManagerListener::UserConnected, const UserPtr& aUser) throw() {
	bool hasDown = false;
	{
		Lock l(cs);
		for(int i = 0; i < QueueItem::LAST; ++i) {
			QueueItem::UserListIter j = userQueue.getList(i).find(aUser);
			if(j != userQueue.getList(i).end()) {
				for(QueueItem::Iter m = j->second.begin(); m != j->second.end(); ++m)
				fire(QueueManagerListener::StatusUpdated(), *m);
				if(i != QueueItem::PAUSED)
					hasDown = true;
			}
		}
	}

	if(hasDown)
		ConnectionManager::getInstance()->getDownloadConnection(aUser);
}

void QueueManager::on(ClientManagerListener::UserDisconnected, const UserPtr& aUser) throw() {
	Lock l(cs);
	for(int i = 0; i < QueueItem::LAST; ++i) {
		QueueItem::UserListIter j = userQueue.getList(i).find(aUser);
		if(j != userQueue.getList(i).end()) {
			for(QueueItem::Iter m = j->second.begin(); m != j->second.end(); ++m)
                fire(QueueManagerListener::StatusUpdated(), *m);
		}
	}
}

void QueueManager::on(TimerManagerListener::Second, uint64_t /*aTick*/) throw() {
	Lock l(cs);
	QueueItem::UserMap& um = userQueue.getRunning();

	for(QueueItem::UserIter j = um.begin(); j != um.end(); ++j) {
		QueueItem* q = j->second;
		dcassert(q->getCurrentDownload() != NULL);
		q->setDownloadedBytes(q->getCurrentDownload()->getPos());
		fire(QueueManagerListener::StatusUpdated(), q);
	}
}

void QueueItem::addSource(const UserPtr& aUser, const string& aPath) {
	dcassert(!isSource(aUser));
	SourceIter i = getBadSource(aUser);
	if(i != badSources.end()) {
        i->setPath(aPath);
		sources.push_back(*i);
		badSources.erase(i);
	} else {
		sources.push_back(Source(aUser, aPath));
	}
}

void QueueItem::removeSource(const UserPtr& aUser, Flags<uint16_t>::MaskType reason) {
	SourceIter i = getSource(aUser);
	dcassert(i != sources.end());
	i->setFlag(reason);
	badSources.push_back(*i);
	sources.erase(i);
}

bool QueueManager::add(const string& aFile, int64_t aSize, const string& tth) throw(QueueException, FileException) {	
	bool newItem = false;
	dcassert(aFile != USER_LIST_NAME);

	string target = SETTING(DOWNLOAD_DIRECTORY) + aFile;
	string tempTarget = SETTING(TEMP_DOWNLOAD_DIRECTORY) + aFile;

	Flags<uint16_t>::MaskType flag = QueueItem::FLAG_RESUME;

	try {
		target = checkTarget(target, aSize, flag);
		if(target.empty())
			return false;
	} catch(const Exception&) {
		return false;
	}

	{
		Lock l(cs);
		QueueItem::List matches;
		TTHValue root(tth);

		fileQueue.find(matches, aSize, Util::getFileExt(aFile));

		if(&root != NULL) {
			fileQueue.find(matches, root);
		}

		for(QueueItem::Iter i = matches.begin(); i != matches.end(); ++i) {
			QueueItem* qi = *i;
			bool found = false;
			if(qi->getTTH()) {
				found = *qi->getTTH() == root;
			} else {
				found = (Util::stricmp(qi->getTargetFileName(), aFile) == 0);
			}

			if(found) {
				return false;
			}
		}

		QueueItem* q = fileQueue.add(target, aSize, flag, QueueItem::DEFAULT, tempTarget, 0, GET_TIME(), &root);
		fire(QueueManagerListener::Added(), q);
		newItem = true;
	}

	if(newItem){
		SearchManager::getInstance()->search(tth, 0, SearchManager::TYPE_TTH, SearchManager::SIZE_DONTCARE);
		return true;
	}
	return false;
}

void QueueManager::clearErrors(const string& aTarget) throw() {
	Lock l(cs);

	QueueItem* q = fileQueue.find(aTarget);
	if((q != NULL)) {
		QueueItem::SourceIter j = q->getBadSources().begin();
		while(j != q->getBadSources().end()) {
			if(!j->isSet(QueueItem::Source::FLAG_REMOVED)) {
				q->badSources.erase(j);
				continue;
			}
			j++;
		}
		setDirty();
		fire(QueueManagerListener::StatusUpdated(), q);
	}
}

string QueueManager::getTempTarget(const string& aTarget) {
	Lock l(cs);

	QueueItem* qi = fileQueue.find(aTarget);
	if(qi == NULL)
		return Util::emptyString;

	return qi->getTempTarget();
}

/*
 * Some parts Copyright (C) 2002-2009 Petr Kozelka, PPK at czdc dot org
 * Licensed under GNU General Public License version 3.
 */
