/*
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef UPLOADMANAGER_H
#define UPLOADMANAGER_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "UserConnection.h"
#include "Singleton.h"

#include "ClientManagerListener.h"
#include "MerkleTree.h"

class InputStream;

class Upload : public Transfer, public Flags<uint8_t> {
public:
	enum Flags {
		FLAG_USER_LIST    = 0x1,
		FLAG_TTH_LEAVES   = 0x2,
		FLAG_ZUPLOAD      = 0x4,
		FLAG_PENDING_KICK = 0x8
	};

	typedef Upload* Ptr;
	typedef vector<Ptr> List;
	typedef List::iterator Iter;

	Upload(UserConnection& conn);
	~Upload();

	void getParams(const UserConnection& aSource, StringMap& params);

	GETSET(string, sourceFile, SourceFile);
	GETSET(InputStream*, stream, Stream);
};

class UploadManagerListener {
	friend class UploadQueueItem;
public:
	virtual ~UploadManagerListener() { }
	template<int I>	struct X { enum { TYPE = I }; };

	typedef X<0> Complete;
	typedef X<1> Failed;
	typedef X<2> Starting;
	typedef X<3> Tick;
	typedef X<4> QueueAdd;
	typedef X<5> QueueRemove;
	typedef X<6> QueueItemRemove;

	virtual void on(Starting, Upload*) throw() { }
	virtual void on(Tick, const Upload::List&) throw() { }
	virtual void on(Complete, Upload*) throw() { }
	virtual void on(Failed, Upload*, const string&) throw() { }
	virtual void on(QueueAdd, UploadQueueItem*) throw() { }
	virtual void on(QueueRemove, const UserPtr&) throw() { }
	virtual void on(QueueItemRemove, UploadQueueItem*) throw() { }
};

class UploadQueueItem : public PointerBase {
	public:
		UploadQueueItem(UserPtr u, string file, int64_t p, int64_t sz, time_t itime) :
			User(u), File(file), pos(p), size(sz), iTime(itime) {
        	inc();
        }
		~UploadQueueItem() throw() { }

		typedef vector<UploadQueueItem*> List;
		typedef List::iterator Iter;

		typedef HASH_MAP<UserPtr, UploadQueueItem::List, User::HashFunction> UserMap;
		typedef UserMap::iterator UserMapIter;

        void update();

		static int compareItems(UploadQueueItem* a, UploadQueueItem* b, int col) {
			switch(col) {
				case COLUMN_TRANSFERRED: return compare(a->pos, b->pos);
				case COLUMN_SIZE: return compare(a->size, b->size);
				case COLUMN_ADDED: return compare(a->iTime, b->iTime);
				case COLUMN_WAITING: return compare(a->iTime, b->iTime);
				default:
					return Util::stricmp(a->columns[col], b->columns[col]);
			}
		}

		enum {
			COLUMN_FIRST,
			COLUMN_FILE = COLUMN_FIRST,
			COLUMN_PATH,
			COLUMN_NICK,
			COLUMN_HUB,
			COLUMN_TRANSFERRED,
			COLUMN_SIZE,
			COLUMN_ADDED,
			COLUMN_WAITING,
			COLUMN_LAST
		};

		const TCHAR* getText(const int &col, TCHAR* buf) const {
            return columns[col].c_str();
		}

		int imageIndex();

		int64_t pos;
		int64_t size;
		time_t iTime;
		UserPtr User;
		string File;
		tstring columns[COLUMN_LAST];
};

class UploadManager : private ClientManagerListener, private UserConnectionListener, public Speaker<UploadManagerListener>,
    private TimerManagerListener, public Singleton<UploadManager> {
public:
	UploadManager() throw();
	virtual ~UploadManager() throw();

	/** @return Number of uploads. */
	size_t getUploadCount() { Lock l(cs); return uploads.size(); }

	/**
	 * @remarks This is only used in the tray icons. Could be used in
	 * MainFrame too.
	 *
	 * @return Running average download speed in Bytes/s
	 */
	int64_t getRunningAverage();

	/** @return Number of free slots. */
	int getFreeSlots() { return max((Util::getSlots() - running), 0); }

	/** @internal */
	int getFreeExtraSlots() { return max(SETTING(EXTRA_SLOTS) - getExtra(), 0); };

	/** @param aUser Reserve an upload slot for this user and connect. */
	void reserveSlot(const UserPtr& aUser);
	void reserveSlotHour(const UserPtr& aUser);
	void reserveSlotDay(const UserPtr& aUser);
	void reserveSlotWeek(const UserPtr& aUser);
	void unreserveSlot(const UserPtr& aUser);

	/** @internal */
	void addConnection(UserConnection::Ptr conn) {
		conn->addListener(this);
		conn->setState(UserConnection::STATE_GET);
	}

	GETSET(int, running, Running);
	GETSET(int, extra, Extra);
	GETSET(uint64_t, lastGrant, LastGrant);

	// Upload throttling
	size_t throttleGetSlice();
	size_t throttleCycleTime();

	UploadQueueItem::UserMap getQueue();
	void clearUserFiles(const UserPtr&);
	UploadQueueItem::UserMap UploadQueueItems;
	static bool getFireballStatus() { return m_boFireball; }
	static bool getFileServerStatus() { return m_boFileServer; }
	bool hasReservedSlot(const UserPtr& aUser) { return reservedSlots.find(aUser) != reservedSlots.end(); }
private:
	void throttleZeroCounters();
	void throttleBytesTransferred(size_t i);
	void throttleSetup();
	bool mThrottleEnable;
	size_t mBytesSent, mBytesSpokenFor, mUploadLimit, mCycleTime, mByteSlice;

	static bool m_boFireball;
	static bool m_boFileServer;

	// Variables for Fireball detecting
	bool m_boLastTickHighSpeed;
	uint64_t m_iHighSpeedStartTick;
	bool boFireballSent;

	// Main fileserver flag
	bool boFileServerSent;

	
	Upload::List uploads;
	CriticalSection cs;

	typedef HASH_MAP<UserPtr, uint64_t, User::HashFunction> SlotMap;
	typedef SlotMap::iterator SlotIter;
	SlotMap reservedSlots;

	void addFailedUpload(UserPtr& User, string file, int64_t pos, int64_t size);

	bool getAutoSlot();
	void removeConnection(UserConnection* aConn);
	void removeUpload(Upload* aUpload);

	// ClientManagerListener
	virtual void on(ClientManagerListener::UserDisconnected, const UserPtr& aUser) throw();

	// TimerManagerListener
	virtual void on(Minute, uint64_t aTick) throw();
	virtual void on(Second, uint64_t aTick) throw();

	// UserConnectionListener
	virtual void on(BytesSent, UserConnection*, size_t, size_t) throw();
	virtual void on(Failed, UserConnection*, const string&) throw();
	virtual void on(Get, UserConnection*, const string&, int64_t) throw();
	virtual void on(GetBlock, UserConnection* conn, const string& line, int64_t resume, int64_t bytes) throw() { onGetBlock(conn, line, resume, bytes, false); }
	virtual void on(Send, UserConnection*) throw();
	virtual void on(GetListLength, UserConnection* conn) throw();
	virtual void on(TransmitDone, UserConnection*) throw();

	virtual void on(AdcCommand::GET, UserConnection*, const AdcCommand&) throw();

	void onGetBlock(UserConnection* aSource, const string& aFile, int64_t aResume, int64_t aBytes, bool z);
	bool prepareFile(UserConnection& aSource, const string& aType, const string& aFile, int64_t aResume, int64_t aBytes);
};

#endif // UPLOADMANAGER_H

/*
 * Some parts Copyright (C) 2002-2009 Petr Kozelka, PPK at czdc dot org
 * Licensed under GNU General Public License version 3.
 */
